import chalk from 'chalk'
import Axios, { AxiosRequestConfig, AxiosResponse } from 'axios'
import { get as pslGet } from 'psl'
import { Cookie, parse as parseSetCookie } from 'set-cookie-parser'

const DEF_TIMEOUT = 10000
const DEF_MAX_RETRIES = 2
const DEF_WITH_SET_COOKIES = false
const DEF_UA = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.80 Safari/537.36'

export const axios = Axios.create({
  headers: {
    'user-agent': DEF_UA
  },
  withCredentials: true
})

export interface RequestConfig extends AxiosRequestConfig {
  timeout?: number
  maxRetries?: number
  withSetCookies?: boolean
}

class CookieController {
  static cacheMap: Map<string, Cookie[]> = new Map()

  private static parseHostToDomains(host: string) {
    const domains = [host, `.${host}`]
    const domain = pslGet(host)

    if (domain && domain !== host) {
      domains.push(domain, `.${domain}`)
    }
    return domains
  }

  static sync(setCookies: string[], host: string) {
    const cookies = parseSetCookie(setCookies)

    for (const item of cookies) {
      console.debug(chalk.white(`└─ set-cookie`, JSON.stringify(item)))
      const { name, domain = host } = item
      const cached = this.cacheMap.get(domain) || []
      const existIndex = cached.findIndex(c => c.name === name)

      if (existIndex > -1) {
        cached.splice(existIndex, 1, item)
      } else {
        cached.push(item)
      }
      this.cacheMap.set(domain, cached)
    }
  }

  static inject(url: string, headers: any) {
    const domains = this.parseHostToDomains(new URL(url).host)
    const cookies: string[] = []
    const now = Date.now()

    for (const domain of domains) {
      const cache = this.cacheMap.get(domain)

      if (cache) {
        const cookieStrs = cache
          .filter(c => !c.expires || c.expires.valueOf() > now)
          .map(c => `${c.name}=${c.value}`)
        cookies.push(...cookieStrs)
      }
    }

    const cookie = (headers.cookie ? `${headers.cookie}; ` : '') + cookies.join('; ')
    headers.cookie = cookie
    cookie && console.debug(chalk.white(`└─ cookie:`, cookie))
  }
}

export const request = (config: RequestConfig, retries = 0): Promise<AxiosResponse<any>> => {
  config.timeout = config.timeout || DEF_TIMEOUT
  config.maxRetries = config.maxRetries || DEF_MAX_RETRIES
  config.withSetCookies = config.withSetCookies != null ? config.withSetCookies : DEF_WITH_SET_COOKIES
  const _config = JSON.stringify(config)
  const fullUrl = `${config.baseURL || axios.defaults.baseURL || ''}${config.url}`
  const cancelTokenSource = Axios.CancelToken.source()
  config.cancelToken = cancelTokenSource.token
  config.headers = config.headers || {}

  console.debug(chalk.white(`[request] ${config.method || 'GET'} ${fullUrl}`, config.params ? JSON.stringify(config.params) : ''))

  if (config.withSetCookies) {
    CookieController.inject(fullUrl, config.headers)
  }

  return new Promise((resolve, reject) => {
    let done = false

    const timer = setTimeout(() => {
      if (!done) {
        cancelTokenSource.cancel()

        // @ts-ignore
        if (retries >= config.maxRetries) {
          reject(new Error(`max retries exceeded (${retries})`))
        } else {
          retries++
          console.info(chalk.magenta(`[request] ${config.url} timeouted, canceled, retry ${retries}`))
          resolve(request(JSON.parse(_config), retries))
        }
      }
    }, config.timeout)

    const startAt = Date.now()

    axios(config)
      .then((res) => {
        console.debug(chalk.white(`└─ ${res.status} ${Date.now() - startAt}ms`))
        const setCookies = res.headers['set-cookie']

        if (config.withSetCookies && setCookies) {
          const { host } = new URL(res.request.res.responseUrl)
          CookieController.sync(setCookies, host)
        }

        resolve(res)
      })
      .catch((e) => {
        if (e.response && e.response.data) {
          console.error('[error.message]', e.message)
          console.error('[error.response]', e.response.data)
        }
        reject(e)
      })
      .finally(() => {
        done = true
        clearTimeout(timer)
      })
  })
}
